@@ -2,8 +2,9 @@
  * vim: set et ts=8 sw=8:
  *
  * Copyright (C) 2008 Novell, Inc.
+ * Copyright (C) 2009 Red Hat, Inc.
  *
- * Authors: Vincent Untz
+ * Authors: Vincent Untz, Tim Waugh
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -58,7 +59,7 @@
      getPPDs
      getServerPPD
      getDocument
-     getDevices
+~!+* getDevices
      getJobs
      getJobAttributes
 ~!+* cancelJob
@@ -230,18 +231,30 @@ cph_cups_new (void)
  * Validation
  ******************************************************/
 
+/* From https://bugzilla.novell.com/show_bug.cgi?id=447444#c5
+ * We need to define a maximum length for strings to avoid cups
+ * thinking there are multiple lines.
+ */
+#define CPH_STR_MAXLEN 512
+
 static gboolean
 _cph_cups_is_string_printable (const char *str,
-                               gboolean    check_for_null)
+                               gboolean    check_for_null,
+                               int         maxlen)
 {
         int i;
+        int len;
 
         /* no NULL string */
         if (!str)
                 return !check_for_null;
 
+        len = strlen (str);
+        if (maxlen > 0 && len > maxlen)
+                return FALSE;
+
         /* only printable characters */
-        for (i = 0; i < strlen (str); i++) {
+        for (i = 0; i < len; i++) {
                 if (!g_ascii_isprint (str[i]))
                         return FALSE;
         }
@@ -249,14 +262,14 @@ _cph_cups_is_string_printable (const char *str,
         return TRUE;
 }
 
-#define _CPH_CUPS_IS_VALID(name, name_for_str, check_for_null)                \
+#define _CPH_CUPS_IS_VALID(name, name_for_str, check_for_null, maxlen)        \
 static gboolean                                                               \
 _cph_cups_is_##name##_valid (CphCups    *cups,                                \
                              const char *str)                                 \
 {                                                                             \
         char *error;                                                          \
                                                                               \
-        if (_cph_cups_is_string_printable (str, check_for_null))              \
+        if (_cph_cups_is_string_printable (str, check_for_null, maxlen))      \
                 return TRUE;                                                  \
                                                                               \
         error = g_strdup_printf ("\"%s\" is not a valid %s.",                 \
@@ -271,13 +284,20 @@ static gboolean
 _cph_cups_is_printer_name_valid_internal (const char *name)
 {
         int i;
+        int len;
 
         /* no empty string */
         if (!name || name[0] == '\0')
                 return FALSE;
 
+        len = strlen (name);
+        /* no string that is too long; see comment at the beginning of the
+         * validation code block */
+        if (len > CPH_STR_MAXLEN)
+                return FALSE;
+
         /* only printable characters, no space, no /, no # */
-        for (i = 0; i < strlen (name); i++) {
+        for (i = 0; i < len; i++) {
                 if (!g_ascii_isprint (name[i]))
                         return FALSE;
                 if (g_ascii_isspace (name[i]))
@@ -289,6 +309,40 @@ _cph_cups_is_printer_name_valid_internal (const char *name)
         return TRUE;
 }
 
+static gboolean
+_cph_cups_is_scheme_valid_internal (const char *scheme)
+{
+        int i;
+        int len;
+
+        /* no empty string */
+        if (!scheme || scheme[0] == '\0')
+                return FALSE;
+
+        len = strlen (scheme);
+        /* no string that is too long; see comment at the beginning of the
+         * validation code block */
+        if (len > CPH_STR_MAXLEN)
+                return FALSE;
+
+        /* From RFC 1738:
+         * Scheme names consist of a sequence of characters. The lower case
+         * letters "a"--"z", digits, and the characters plus ("+"), period
+         * ("."), and hyphen ("-") are allowed. For resiliency, programs
+         * interpreting URLs should treat upper case letters as equivalent to
+         * lower case in scheme names (e.g., allow "HTTP" as well as "http").
+         */
+        for (i = 0; i < len; i++) {
+                if (!g_ascii_isalnum (scheme[i]) &&
+                    scheme[i] != '+' &&
+                    scheme[i] != '.' &&
+                    scheme[i] != '-')
+                        return FALSE;
+        }
+
+        return TRUE;
+}
+
 static gboolean
 _cph_cups_is_printer_name_valid (CphCups    *cups,
                                  const char *name)
@@ -338,6 +392,22 @@ _cph_cups_is_job_id_valid (CphCups *cups,
         return FALSE;
 }
 
+static gboolean
+_cph_cups_is_scheme_valid (CphCups    *cups,
+                           const char *scheme)
+{
+        char *error;
+
+        if (_cph_cups_is_scheme_valid_internal (scheme))
+                return TRUE;
+
+        error = g_strdup_printf ("\"%s\" is not a valid scheme.", scheme);
+        _cph_cups_set_internal_status (cups, error);
+        g_free (error);
+
+        return FALSE;
+}
+
 /* This is some text, but we could potentially do more checks. We don't do them
  * because cups will already do them.
  *   + for the URI, we could check that the scheme is supported and that the
@@ -353,12 +423,12 @@ _cph_cups_is_job_id_valid (CphCups *cups,
  *     printer-error-policy-supported and printer-op-policy-supported
  *     attributes.
  */
-_CPH_CUPS_IS_VALID (printer_uri, "printer URI", TRUE)
-_CPH_CUPS_IS_VALID (ppd, "PPD", TRUE)
-_CPH_CUPS_IS_VALID (ppd_filename, "PPD file", FALSE)
-_CPH_CUPS_IS_VALID (job_sheet, "job sheet", FALSE)
-_CPH_CUPS_IS_VALID (error_policy, "error policy", FALSE)
-_CPH_CUPS_IS_VALID (op_policy, "op policy", FALSE)
+_CPH_CUPS_IS_VALID (printer_uri, "printer URI", TRUE, CPH_STR_MAXLEN)
+_CPH_CUPS_IS_VALID (ppd, "PPD", TRUE, CPH_STR_MAXLEN)
+_CPH_CUPS_IS_VALID (ppd_filename, "PPD file", FALSE, CPH_STR_MAXLEN)
+_CPH_CUPS_IS_VALID (job_sheet, "job sheet", FALSE, CPH_STR_MAXLEN)
+_CPH_CUPS_IS_VALID (error_policy, "error policy", FALSE, CPH_STR_MAXLEN)
+_CPH_CUPS_IS_VALID (op_policy, "op policy", FALSE, CPH_STR_MAXLEN)
 
 /* Check for users. Those are some printable strings, which souldn't be NULL.
  * They should also not be empty, but it appears that it's possible to carry
@@ -367,7 +437,7 @@ _CPH_CUPS_IS_VALID (op_policy, "op policy", FALSE)
  * We could also check that the username exists on the system, but cups will do
  * it.
  */
-_CPH_CUPS_IS_VALID (user, "user", TRUE)
+_CPH_CUPS_IS_VALID (user, "user", TRUE, CPH_STR_MAXLEN)
 
 /* Check for options & values. Those are for sure some printable strings, but
  * can we do more? Let's see:
@@ -376,14 +446,14 @@ _CPH_CUPS_IS_VALID (user, "user", TRUE)
  *     and so we'll let cups handle that.
  *   + a value can be some text, and we don't know much more.
  */
-_CPH_CUPS_IS_VALID (option, "option", TRUE)
-_CPH_CUPS_IS_VALID (option_value, "value for option", FALSE)
+_CPH_CUPS_IS_VALID (option, "option", TRUE, CPH_STR_MAXLEN)
+_CPH_CUPS_IS_VALID (option_value, "value for option", FALSE, CPH_STR_MAXLEN)
 
 /* This is really just some text */
-_CPH_CUPS_IS_VALID (info, "description", FALSE)
-_CPH_CUPS_IS_VALID (location, "location", FALSE)
-_CPH_CUPS_IS_VALID (reject_jobs_reason, "reason", FALSE)
-_CPH_CUPS_IS_VALID (job_hold_until, "job hold until", FALSE)
+_CPH_CUPS_IS_VALID (info, "description", FALSE, CPH_STR_MAXLEN)
+_CPH_CUPS_IS_VALID (location, "location", FALSE, CPH_STR_MAXLEN)
+_CPH_CUPS_IS_VALID (reject_jobs_reason, "reason", FALSE, CPH_STR_MAXLEN)
+_CPH_CUPS_IS_VALID (job_hold_until, "job hold until", FALSE, CPH_STR_MAXLEN)
 
 /* For put/get file: this is some text, but we could potentially do more
  * checks. We don't do them because cups will already do them.
@@ -391,8 +461,8 @@ _CPH_CUPS_IS_VALID (job_hold_until, "job hold until", FALSE)
  *   + for the filename, in the put case, we could check that the file exists
  *     and is a regular file (no socket, block device, etc.).
  */
-_CPH_CUPS_IS_VALID (resource, "resource", TRUE)
-_CPH_CUPS_IS_VALID (filename, "filename", TRUE)
+_CPH_CUPS_IS_VALID (resource, "resource", TRUE, CPH_STR_MAXLEN)
+_CPH_CUPS_IS_VALID (filename, "filename", TRUE, CPH_STR_MAXLEN)
 
 /******************************************************
  * Helpers
@@ -458,25 +528,37 @@ _cph_cups_set_error_from_reply (CphCups *cups,
 }
 
 static gboolean
-_cph_cups_handle_reply (CphCups *cups,
-                        ipp_t   *reply)
+_cph_cups_is_reply_ok (CphCups  *cups,
+                       ipp_t    *reply,
+                       gboolean  delete_reply_if_not_ok)
 {
-        gboolean retval;
-
         /* reset the internal status: we'll use the cups status */
         _cph_cups_set_internal_status (cups, NULL);
 
-        if (!reply || reply->request.status.status_code > IPP_OK_CONFLICT) {
-                retval = FALSE;
+        if (reply && reply->request.status.status_code <= IPP_OK_CONFLICT) {
+                cups->priv->last_status = IPP_OK;
+                return TRUE;
+        } else {
                 _cph_cups_set_error_from_reply (cups, reply);
 #if 0
                 /* Useful when debugging: */
                 g_print ("%s\n", cupsLastErrorString ());
 #endif
-        } else {
-                retval = TRUE;
-                cups->priv->last_status = IPP_OK;
+
+                if (delete_reply_if_not_ok && reply)
+                        ippDelete (reply);
+
+                return FALSE;
         }
+}
+
+static gboolean
+_cph_cups_handle_reply (CphCups *cups,
+                        ipp_t   *reply)
+{
+        gboolean retval;
+
+        retval = _cph_cups_is_reply_ok (cups, reply, FALSE);
 
         if (reply)
                 ippDelete (reply);
@@ -639,6 +721,38 @@ _cph_cups_send_new_job_attributes_request (CphCups     *cups,
         return _cph_cups_send_request (cups, request, resource);
 }
 
+static const char *
+_cph_cups_get_attribute_string (ipp_attribute_t *attrs,
+                                ipp_tag_t        group,
+                                const char      *name,
+                                ipp_tag_t        type)
+{
+        ipp_attribute_t *attr;
+
+        for (attr = attrs; attr; attr = attr->next) {
+                while (attr && attr->group_tag != group)
+                        attr = attr->next;
+
+                if (attr == NULL)
+                        break;
+
+                while (attr && attr->group_tag == group) {
+                        if (attr->name &&
+                            strcmp (attr->name, name) == 0 &&
+                            attr->value_tag == type) {
+                                return attr->values[0].string.text;
+                        }
+
+                        attr = attr->next;
+                }
+
+                if (attr == NULL)
+                        break;
+        }
+
+        return NULL;
+}
+
 static int
 _cph_cups_class_has_printer (CphCups     *cups,
                              const char  *class_name,
@@ -796,7 +910,7 @@ cph_cups_is_class (CphCups    *cups,
         reply = cupsDoRequest (cups->priv->connection,
                                request, resource_char);
 
-        if (!reply)
+        if (!_cph_cups_is_reply_ok (cups, reply, TRUE))
                 return FALSE;
 
         /* Note: we need to look if the attribute is there, since we get a
@@ -818,7 +932,7 @@ cph_cups_printer_get_uri (CphCups    *cups,
         ipp_t              *request;
         const char         *resource_char;
         ipp_t              *reply;
-        ipp_attribute_t    *attr;
+        const char         *const_uri;
         char               *uri;
 
         g_return_val_if_fail (CPH_IS_CUPS (cups), NULL);
@@ -835,32 +949,16 @@ cph_cups_printer_get_uri (CphCups    *cups,
         reply = cupsDoRequest (cups->priv->connection,
                                request, resource_char);
 
-        if (!reply)
+        if (!_cph_cups_is_reply_ok (cups, reply, TRUE))
                 return NULL;
 
+        const_uri = _cph_cups_get_attribute_string (reply->attrs, IPP_TAG_PRINTER,
+                                                    attrs[0], IPP_TAG_URI);
+
         uri = NULL;
 
-        for (attr = reply->attrs; attr; attr = attr->next) {
-                while (attr && attr->group_tag != IPP_TAG_PRINTER)
-                        attr = attr->next;
-
-                if (attr == NULL)
-                        break;
-
-                while (attr && attr->group_tag == IPP_TAG_PRINTER) {
-                        if (attr->name &&
-                            strcmp (attr->name, attrs[0]) == 0 &&
-                            attr->value_tag == IPP_TAG_URI) {
-                                uri = g_strdup (attr->values[0].string.text);
-                                break;
-                        }
-
-                        attr = attr->next;
-                }
-
-                if (uri != NULL || attr == NULL)
-                        break;
-        }
+        if (const_uri)
+                uri = g_strdup (const_uri);
 
         ippDelete (reply);
 
@@ -967,6 +1065,369 @@ cph_cups_file_put (CphCups    *cups,
                 cups->priv->last_status == HTTP_CREATED);
 }
 
+/* Functions that are for the server in general */
+
+GHashTable *
+cph_cups_server_get_settings (CphCups *cups)
+{
+        int            retval;
+        GHashTable    *hash;
+        cups_option_t *settings;
+        int            num_settings, i;
+
+        g_return_val_if_fail (CPH_IS_CUPS (cups), NULL);
+
+        retval = cupsAdminGetServerSettings (cups->priv->connection,
+                                             &num_settings, &settings);
+
+        if (retval == 0) {
+                _cph_cups_set_internal_status (cups,
+                                               "Cannot get server settings.");
+
+                return NULL;
+        }
+
+        hash = g_hash_table_new_full (g_str_hash, g_str_equal,
+                                      g_free, g_free);
+
+        for (i = 0; i < num_settings; i++)
+                g_hash_table_replace (hash,
+                                      g_strdup (settings[i].name),
+                                      g_strdup (settings[i].value));
+
+        cupsFreeOptions (num_settings, settings);
+
+        return hash;
+}
+
+gboolean
+cph_cups_server_set_settings (CphCups    *cups,
+                              GHashTable *settings)
+{
+        int             retval;
+        GHashTableIter  iter;
+        /* key and value are strings, but we want to avoid compiler warnings */
+        gpointer        key;
+        gpointer        value;
+        cups_option_t  *cups_settings;
+        int             num_settings;
+
+        g_return_val_if_fail (CPH_IS_CUPS (cups), FALSE);
+        g_return_val_if_fail (settings != NULL, FALSE);
+
+        /* First pass to check the validity of the hashtable content */
+        g_hash_table_iter_init (&iter, settings);
+        while (g_hash_table_iter_next (&iter, &key, &value)) {
+                if (!_cph_cups_is_option_valid (cups, key))
+                        return FALSE;
+                if (!_cph_cups_is_option_value_valid (cups, value))
+                        return FALSE;
+        }
+
+        /* Second pass to actually set the settings */
+        cups_settings = NULL;
+        num_settings = 0;
+
+        g_hash_table_iter_init (&iter, settings);
+        while (g_hash_table_iter_next (&iter, &key, &value))
+                num_settings = cupsAddOption (key, value,
+                                              num_settings, &cups_settings);
+
+        retval = cupsAdminSetServerSettings (cups->priv->connection,
+                                             num_settings, cups_settings);
+
+        /* CUPS is being restarted, so we need to reconnect */
+        cph_cups_reconnect (cups);
+
+        cupsFreeOptions (num_settings, cups_settings);
+
+        if (retval == 0) {
+                _cph_cups_set_internal_status (cups,
+                                               "Cannot set server settings.");
+
+                return FALSE;
+        }
+
+        return TRUE;
+}
+
+typedef struct {
+        int         iter;
+        int         limit;
+        GHashTable *hash;
+} CphCupsGetDevices;
+
+static void
+_cph_cups_get_devices_cb (const char *device_class,
+                          const char *device_id,
+                          const char *device_info,
+                          const char *device_make_and_model,
+                          const char *device_uri,
+                          const char *device_location,
+                          void       *user_data)
+{
+        CphCupsGetDevices *data = user_data;
+
+        g_return_if_fail (data != NULL);
+
+        if (data->limit > 0 && data->iter >= data->limit)
+                return;
+
+        if (device_class && device_class[0] != '\0')
+                g_hash_table_replace (data->hash,
+                                      g_strdup_printf ("device-class:%d",
+                                                       data->iter),
+                                      g_strdup (device_class));
+        if (device_id && device_id[0] != '\0')
+                g_hash_table_replace (data->hash,
+                                      g_strdup_printf ("device-id:%d",
+                                                       data->iter),
+                                      g_strdup (device_id));
+        if (device_info && device_info[0] != '\0')
+                g_hash_table_replace (data->hash,
+                                      g_strdup_printf ("device-info:%d",
+                                                       data->iter),
+                                      g_strdup (device_info));
+        if (device_make_and_model && device_make_and_model[0] != '\0')
+                g_hash_table_replace (data->hash,
+                                      g_strdup_printf ("device-make-and-model:%d",
+                                                       data->iter),
+                                      g_strdup (device_make_and_model));
+        if (device_uri && device_uri[0] != '\0')
+                g_hash_table_replace (data->hash,
+                                      g_strdup_printf ("device-uri:%d",
+                                                       data->iter),
+                                      g_strdup (device_uri));
+        if (device_location && device_location[0] != '\0')
+                g_hash_table_replace (data->hash,
+                                      g_strdup_printf ("device-location:%d ",
+                                                       data->iter),
+                                      g_strdup (device_location));
+
+        data->iter++;
+}
+
+#if (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR >= 4) || CUPS_VERSION_MAJOR > 1
+static gboolean
+_cph_cups_devices_get_14 (CphCups            *cups,
+                          int                 timeout,
+                          int                 limit,
+                          const char        **include_schemes,
+                          const char        **exclude_schemes,
+                          int                 len_include,
+                          int                 len_exclude,
+                          CphCupsGetDevices  *data)
+{
+        ipp_status_t  retval;
+        int           timeout_param = CUPS_TIMEOUT_DEFAULT;
+        char         *include_schemes_param;
+        char         *exclude_schemes_param;
+
+        if (timeout > 0)
+                timeout_param = timeout;
+
+        if (include_schemes && len_include > 0)
+                include_schemes_param = g_strjoin (",", include_schemes);
+        else
+                include_schemes_param = g_strdup (CUPS_INCLUDE_ALL);
+
+        if (exclude_schemes && len_exclude > 0)
+                exclude_schemes_param = g_strjoin (",", exclude_schemes);
+        else
+                exclude_schemes_param = g_strdup (CUPS_EXCLUDE_NONE);
+
+        retval = cupsGetDevices (cups->priv->connection,
+                                 timeout_param,
+                                 include_schemes_param,
+                                 exclude_schemes_param,
+                                 _cph_cups_get_devices_cb,
+                                 data);
+
+        g_free (include_schemes_param);
+        g_free (exclude_schemes_param);
+
+        if (retval != IPP_OK) {
+                _cph_cups_set_internal_status (cups,
+                                               "Cannot get devices.");
+                return FALSE;
+        }
+
+        return TRUE;
+}
+#else
+static gboolean
+_cph_cups_devices_get_old (CphCups            *cups,
+                           int                 timeout,
+                           int                 limit,
+                           const char        **include_schemes,
+                           const char        **exclude_schemes,
+                           int                 len_include,
+                           int                 len_exclude,
+                           CphCupsGetDevices  *data)
+{
+        ipp_t           *request;
+        const char      *resource_char;
+        ipp_t           *reply;
+        ipp_attribute_t *attr;
+        const char      *device_class;
+        const char      *device_id;
+        const char      *device_info;
+        const char      *device_location;
+        const char      *device_make_and_model;
+        const char      *device_uri;
+
+        request = ippNewRequest (CUPS_GET_DEVICES);
+
+        if (timeout > 0)
+                ippAddInteger (request, IPP_TAG_OPERATION, IPP_TAG_INTEGER,
+                               "timeout", timeout);
+        if (limit > 0)
+                ippAddInteger (request, IPP_TAG_OPERATION, IPP_TAG_INTEGER,
+                               "limit", limit);
+
+        if (include_schemes && len_include > 0) {
+                int i;
+
+                attr = ippAddStrings (request, IPP_TAG_OPERATION, IPP_TAG_NAME,
+                                      "include-schemes", len_include, NULL, NULL);
+                for (i = 0; i < len_include; i++)
+                        attr->values[i].string.text = g_strdup (include_schemes[i]);
+        }
+
+        if (exclude_schemes && len_exclude > 0) {
+                int i;
+
+                attr = ippAddStrings (request, IPP_TAG_OPERATION, IPP_TAG_NAME,
+                                      "exclude-schemes", len_exclude, NULL, NULL);
+                for (i = 0; i < len_exclude; i++)
+                        attr->values[i].string.text = g_strdup (exclude_schemes[i]);
+        }
+
+        resource_char = _cph_cups_get_resource (CPH_RESOURCE_ROOT);
+        reply = cupsDoRequest (cups->priv->connection,
+                               request, resource_char);
+
+        if (!_cph_cups_is_reply_ok (cups, reply, TRUE))
+                return FALSE;
+
+        for (attr = reply->attrs; attr; attr = attr->next) {
+                while (attr && attr->group_tag != IPP_TAG_PRINTER)
+                        attr = attr->next;
+
+                if (attr == NULL)
+                        break;
+
+                device_class          = NULL;
+                device_id             = NULL;
+                device_info           = NULL;
+                device_location       = NULL;
+                device_make_and_model = NULL;
+                device_uri            = NULL;
+
+                while (attr && attr->group_tag == IPP_TAG_PRINTER) {
+                        if (attr->name == NULL)
+                                /* nothing, just skip */;
+                        else if (strcmp (attr->name, "device-class") == 0 &&
+                                 attr->value_tag == IPP_TAG_KEYWORD)
+                                device_class = g_strdup (attr->values[0].string.text);
+                        else if (strcmp (attr->name, "device-id") == 0 &&
+                                 attr->value_tag == IPP_TAG_TEXT)
+                                device_id = g_strdup (attr->values[0].string.text);
+                        else if (strcmp (attr->name, "device-info") == 0 &&
+                                 attr->value_tag == IPP_TAG_TEXT)
+                                device_info = g_strdup (attr->values[0].string.text);
+                        else if (strcmp (attr->name, "device-location") == 0 &&
+                                 attr->value_tag == IPP_TAG_TEXT)
+                                device_location = g_strdup (attr->values[0].string.text);
+                        else if (strcmp (attr->name, "device-make-and-model") == 0 &&
+                                 attr->value_tag == IPP_TAG_TEXT)
+                                device_make_and_model = g_strdup (attr->values[0].string.text);
+                        else if (strcmp (attr->name, "device-uri") == 0 &&
+                                 attr->value_tag == IPP_TAG_URI)
+                                device_uri = g_strdup (attr->values[0].string.text);
+
+                        attr = attr->next;
+                }
+
+                if (device_uri)
+                        _cph_cups_get_devices_cb (device_class,
+                                                  device_id,
+                                                  device_info,
+                                                  device_make_and_model,
+                                                  device_uri,
+                                                  device_location,
+                                                  data);
+
+                if (attr == NULL)
+                        break;
+        }
+
+        ippDelete (reply);
+
+        return TRUE;
+}
+#endif
+
+GHashTable *
+cph_cups_devices_get (CphCups     *cups,
+                      int          timeout,
+                      int          limit,
+                      const char **include_schemes,
+                      const char **exclude_schemes)
+{
+        CphCupsGetDevices data;
+        int               len_include;
+        int               len_exclude;
+        gboolean          retval;
+
+        g_return_val_if_fail (CPH_IS_CUPS (cups), NULL);
+
+        /* check the validity of values */
+        len_include = 0;
+        if (include_schemes) {
+                while (include_schemes[len_include] != NULL) {
+                        if (!_cph_cups_is_scheme_valid (cups, include_schemes[len_include]))
+                                return NULL;
+                        len_include++;
+                }
+        }
+
+        len_exclude = 0;
+        if (exclude_schemes) {
+                while (exclude_schemes[len_exclude] != NULL) {
+                        if (!_cph_cups_is_scheme_valid (cups, exclude_schemes[len_exclude]))
+                                return NULL;
+                        len_exclude++;
+                }
+        }
+
+        data.iter  = 0;
+        data.limit = -1;
+        data.hash  = g_hash_table_new_full (g_str_hash, g_str_equal,
+                                            g_free, g_free);
+        if (limit > 0)
+                data.limit = limit;
+
+#if (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR >= 4) || CUPS_VERSION_MAJOR > 1
+        retval = _cph_cups_devices_get_14 (cups, timeout, limit,
+                                           include_schemes, exclude_schemes,
+                                           len_include, len_exclude,
+                                           &data);
+#else
+        retval = _cph_cups_devices_get_old (cups, timeout, limit,
+                                            include_schemes, exclude_schemes,
+                                            len_include, len_exclude,
+                                            &data);
+#endif
+
+        if (retval)
+                return data.hash;
+        else {
+                g_hash_table_destroy (data.hash);
+                return NULL;
+        }
+}
+
 /* Functions that work on a printer */
 
 gboolean
@@ -997,11 +1458,13 @@ cph_cups_printer_add (CphCups    *cups,
 
         ippAddString (request, IPP_TAG_PRINTER, IPP_TAG_NAME,
                       "printer-name", NULL, printer_name);
-        ippAddString (request, IPP_TAG_PRINTER, IPP_TAG_URI,
-                      "device-uri", NULL, printer_uri);
         ippAddString (request, IPP_TAG_PRINTER, IPP_TAG_NAME,
                       "ppd-name", NULL, ppd_file);
 
+        if (printer_uri && printer_uri[0] != '\0') {
+                ippAddString (request, IPP_TAG_PRINTER, IPP_TAG_URI,
+                              "device-uri", NULL, printer_uri);
+        }
         if (info && info[0] != '\0') {
                 ippAddString (request, IPP_TAG_PRINTER, IPP_TAG_TEXT,
                               "printer-info", NULL, info);
@@ -1624,105 +2087,16 @@ out:
         return retval;
 }
 
-/* Functions that are for the server in general */
-
-GHashTable *
-cph_cups_server_get_settings (CphCups *cups)
-{
-        int            retval;
-        GHashTable    *hash;
-        cups_option_t *settings;
-        int            num_settings, i;
-
-        g_return_val_if_fail (CPH_IS_CUPS (cups), NULL);
-
-        retval = cupsAdminGetServerSettings (cups->priv->connection,
-                                             &num_settings, &settings);
-
-        if (retval == 0) {
-                char *error;
-
-                error = g_strdup_printf ("Can not get server settings.");
-                _cph_cups_set_internal_status (cups, error);
-                g_free (error);
-
-                return NULL;
-        }
-
-        hash = g_hash_table_new_full (g_str_hash, g_str_equal,
-                                      g_free, g_free);
-
-        for (i = 0; i < num_settings; i++)
-                g_hash_table_replace (hash,
-                                      g_strdup (settings[i].name),
-                                      g_strdup (settings[i].value));
-
-        cupsFreeOptions (num_settings, settings);
-
-        return hash;
-}
-
-gboolean
-cph_cups_server_set_settings (CphCups    *cups,
-                              GHashTable *settings)
-{
-        int             retval;
-        GHashTableIter  iter;
-        /* key and value are strings, but we want to avoid compiler warnings */
-        gpointer        key;
-        gpointer        value;
-        cups_option_t  *cups_settings;
-        int             num_settings;
-
-        g_return_val_if_fail (CPH_IS_CUPS (cups), FALSE);
-        g_return_val_if_fail (settings != NULL, FALSE);
-
-        /* First pass to check the validity of the hashtable content */
-        g_hash_table_iter_init (&iter, settings);
-        while (g_hash_table_iter_next (&iter, &key, &value)) {
-                if (!_cph_cups_is_option_valid (cups, key))
-                        return FALSE;
-                if (!_cph_cups_is_option_value_valid (cups, value))
-                        return FALSE;
-        }
-
-        /* Second pass to actually set the settings */
-        cups_settings = NULL;
-        num_settings = 0;
-
-        g_hash_table_iter_init (&iter, settings);
-        while (g_hash_table_iter_next (&iter, &key, &value))
-                num_settings = cupsAddOption (key, value,
-                                              num_settings, &cups_settings);
-
-        retval = cupsAdminSetServerSettings (cups->priv->connection,
-                                             num_settings, cups_settings);
-
-        /* CUPS is being restarted, so we need to reconnect */
-        cph_cups_reconnect (cups);
-
-        cupsFreeOptions (num_settings, cups_settings);
-
-        if (retval == 0) {
-                char *error;
-
-                error = g_strdup_printf ("Can not set server settings.");
-                _cph_cups_set_internal_status (cups, error);
-                g_free (error);
-
-                return FALSE;
-        }
-
-        return TRUE;
-}
-
 /* Functions that work on jobs */
 
 gboolean
 cph_cups_job_cancel (CphCups    *cups,
                      int         job_id,
+                     gboolean    purge_job,
                      const char *user_name)
 {
+        ipp_t *request;
+
         g_return_val_if_fail (CPH_IS_CUPS (cups), FALSE);
 
         if (!_cph_cups_is_job_id_valid (cups, job_id))
@@ -1730,10 +2104,19 @@ cph_cups_job_cancel (CphCups    *cups,
         /* we don't check if the user name is valid or not because it comes
          * from getpwuid(), and not dbus */
 
-        return _cph_cups_send_new_simple_job_request (cups, IPP_CANCEL_JOB,
-                                                      job_id,
-                                                      user_name,
-                                                      CPH_RESOURCE_JOBS);
+        request = ippNewRequest (IPP_CANCEL_JOB);
+        _cph_cups_add_job_uri (request, job_id);
+
+        if (user_name != NULL)
+                ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_NAME,
+                              "requesting-user-name", NULL, user_name);
+
+#if (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR >= 4) || CUPS_VERSION_MAJOR > 1
+        if (purge_job)
+                ippAddBoolean (request, IPP_TAG_OPERATION, "purge-job", 1);
+#endif
+
+        return _cph_cups_send_request (cups, request, CPH_RESOURCE_JOBS);
 }
 
 gboolean
@@ -1782,27 +2165,43 @@ cph_cups_job_get_status (CphCups    *cups,
                          int         job_id,
                          const char *user)
 {
-        CphJobStatus  status = CPH_JOB_STATUS_INVALID;
-        cups_job_t   *jobs;
-        int           num_jobs;
-        int           i;
+        const char * const  attrs[1] = { "job-originating-user-name" };
+        ipp_t              *request;
+        const char         *resource_char;
+        ipp_t              *reply;
+        const char         *orig_user;
+        CphJobStatus        status;
 
         g_return_val_if_fail (CPH_IS_CUPS (cups), CPH_JOB_STATUS_INVALID);
 
-        num_jobs = cupsGetJobs2 (cups->priv->connection, &jobs, NULL, 0, 0);
+        if (!_cph_cups_is_job_id_valid (cups, job_id))
+                return CPH_JOB_STATUS_INVALID;
 
-        for (i = 0; i < num_jobs; i++) {
-                if (jobs[i].id == job_id) {
-                        if (user != NULL &&
-                            g_strcmp0 (jobs[i].user, user) == 0)
-                                status = CPH_JOB_STATUS_OWNED_BY_USER;
-                        else
-                                status = CPH_JOB_STATUS_NOT_OWNED_BY_USER;
-                        break;
-                }
+        request = ippNewRequest (IPP_GET_JOB_ATTRIBUTES);
+        _cph_cups_add_job_uri (request, job_id);
+        ippAddStrings (request, IPP_TAG_OPERATION, IPP_TAG_KEYWORD,
+                       "requested-attributes", 1, NULL, attrs);
+
+        resource_char = _cph_cups_get_resource (CPH_RESOURCE_ROOT);
+        reply = cupsDoRequest (cups->priv->connection,
+                               request, resource_char);
+
+        if (!_cph_cups_is_reply_ok (cups, reply, TRUE))
+                return CPH_JOB_STATUS_INVALID;
+
+        orig_user = _cph_cups_get_attribute_string (reply->attrs, IPP_TAG_JOB,
+                                                    attrs[0], IPP_TAG_NAME);
+
+        status = CPH_JOB_STATUS_INVALID;
+
+        if (orig_user) {
+                if (g_strcmp0 (orig_user, user) == 0)
+                        status = CPH_JOB_STATUS_OWNED_BY_USER;
+                else
+                        status = CPH_JOB_STATUS_NOT_OWNED_BY_USER;
         }
 
-        cupsFreeJobs (num_jobs, jobs);
+        ippDelete (reply);
 
         return status;
 }
